from typing import Dict, Optional, Tuple, Union, List, Any

import requests

from pycarlo.common import settings

_DEFAULT_DBT_CLOUD_BASE_URL = 'https://cloud.getdbt.com/api/v2'


class DbtCloudClient:
    """
    Simple client for dbt cloud to retrieve projects/jobs/runs
    """
    def __init__(self,
                 dbt_cloud_api_token: Optional[str] = None,
                 dbt_cloud_account_id: Optional[str] = None,
                 dbt_cloud_base_url: str = _DEFAULT_DBT_CLOUD_BASE_URL):
        self._dbt_cloud_api_token = dbt_cloud_api_token or settings.DBT_CLOUD_API_TOKEN
        self._dbt_cloud_account_id = dbt_cloud_account_id or settings.DBT_CLOUD_ACCOUNT_ID
        if not self._dbt_cloud_api_token or not self._dbt_cloud_account_id:
            raise ValueError('Must set DBT_CLOUD_API_TOKEN and DBT_CLOUD_ACCOUNT_ID environment variables!')
        self._dbt_cloud_api_endpoint = f'{dbt_cloud_base_url}/accounts/{self._dbt_cloud_account_id}'

    def get_projects(self) -> Tuple[Dict, List[Dict]]:
        """
        Get all dbt cloud projects in account. See https://docs.getdbt.com/dbt-cloud/api-v2#operation/listProjects

        :return: tuple of (response status (dict), project (List[dict]))
        """
        return self._get('/projects/')

    def get_project(self, project_id: str) -> Tuple[Dict, Dict]:
        """
        Get a single dbt project. See https://docs.getdbt.com/dbt-cloud/api-v2#operation/getProjectById

        :param project_id: dbt cloud project id
        :return: tuple of (response status (dict), project (dict))
        """
        return self._get(f'/projects/{project_id}')

    def get_jobs(self, project_id: str) -> Tuple[Dict, List[Dict]]:
        """
        Get jobs within a project. See https://docs.getdbt.com/dbt-cloud/api-v2#operation/listJobsForAccount

        :param project_id: dbt cloud project id
        :return: tuple of (response status (dict), jobs (List[dict]))
        """
        return self._get('/jobs/', params=dict(
            project_id=project_id
        ))

    def get_job(self, job_id: str) -> Tuple[Dict, Dict]:
        """
        Get a single job. See https://docs.getdbt.com/dbt-cloud/api-v2#operation/getJobById

        :param job_id: dbt cloud job id
        :return: tuple of (response status (dict), job (dict))
        """
        return self._get(f'/jobs/{job_id}')

    def get_runs(self,
                 limit: int = 5,
                 order_by: str = '-finished_at',
                 job_definition_id: Optional[str] = None,
                 include_related: str = '["run_steps","debug_logs"]') -> Tuple[Dict, Dict]:
        """
        Get runs. See https://docs.getdbt.com/dbt-cloud/api-v2#tag/Runs

        :param limit: Max number of runs to return
        :param order_by: Which field to order by
        :param job_definition_id: dbt cloud job id to filter
        :param include_related: include related objects

        :return: tuple of (response status (dict), runs (List[dict]))
        """
        params = dict(
            order_by=order_by,
            limit=limit,
            include_related=include_related
        )
        if job_definition_id:
            params['job_definition_id'] = job_definition_id
        return self._get('/runs/', params=params)

    def get_run_artifact(self, run_id: int, artifact_path: str) -> Dict:
        """
        Get run artifact. See https://docs.getdbt.com/dbt-cloud/api-v2#operation/getArtifactsByRunId

        :param run_id: dbt cloud run id
        :param artifact_path: Artifact path
        :return: artifact (dict)
        """
        return self._get(f'/runs/{run_id}/artifacts/{artifact_path}', return_payload=True)

    def _get(self, path: str, params: Dict = {}, return_payload: bool = False) -> Union[Tuple[Dict, Any], Dict]:
        response = requests.get(
            f'{self._dbt_cloud_api_endpoint}{path}',
            params=params,
            headers={
                'Authorization': f'Token {self._dbt_cloud_api_token}'
            })
        response.raise_for_status()
        payload = response.json()
        if return_payload:
            return payload
        return payload['status'], payload['data']


from typing import Dict


class GqlError(Exception):
    def __init__(self, body=None, headers: Dict = None, message: str = '', status_code: int = 0, summary: str = ''):
        self.body = body
        self.headers = headers
        self.message = message
        self.status_code = status_code
        self.summary = summary
        self.retryable = status_code >= 500
        super(GqlError, self).__init__()

    def __str__(self) -> str:
        return self.summary


class InvalidSessionError(Exception):
    pass


class InvalidConfigFileError(Exception):
    pass



IMPORT_DBT_MANIFEST = """
mutation importDbtManifest($dbtSchemaVersion: String!, $manifestNodesJson: String!, $projectName: String, $defaultResource: String) {
  importDbtManifest(
    dbtSchemaVersion: $dbtSchemaVersion,
    manifestNodesJson: $manifestNodesJson,
    projectName: $projectName,
    defaultResource: $defaultResource
  ) {
    response {
      nodeIdsImported
    }
  }
}
"""

IMPORT_DBT_RUN_RESULTS = """
mutation importDbtRunResults($dbtSchemaVersion: String!, $runResultsJson: String!, $projectName: String, $runId: String, $runLogs: String) {
  importDbtRunResults(
    dbtSchemaVersion: $dbtSchemaVersion,
    runResultsJson: $runResultsJson,
    projectName: $projectName,
    runId: $runId,
    runLogs: $runLogs
  ) {
    response {
      numResultsImported
    }
  }
}
"""

CREATE_PROJECT = """
mutation createDbtProject($projectName: String!, $source: DbtProjectSource!) {
  createDbtProject(
    projectName: $projectName,
    source: $source
  ) {
    dbtProject {
      projectName
      source
    }
  }
}
"""


import uuid
from typing import Optional, Union, Dict

from box import Box, BoxList

from pycarlo.common import get_logger, settings
from pycarlo.common.retries import Backoff, ExponentialBackoffJitter
from pycarlo.common.settings import DEFAULT_MCD_API_ID_HEADER, DEFAULT_MCD_API_TOKEN_HEADER, \
    DEFAULT_RETRY_INITIAL_WAIT_TIME, DEFAULT_RETRY_MAX_WAIT_TIME, \
    DEFAULT_MCD_SESSION_ID, DEFAULT_MCD_TRACE_ID, DEFAULT_MCD_USER_ID_HEADER
from pycarlo.core.endpoint import Endpoint
from pycarlo.core.operations import Query, Mutation
from pycarlo.core.session import Session

logger = get_logger(__name__)


class Client:
    def __init__(self, session: Optional[Session] = None):
        """
        Create a client for making requests to the MCD API.

        :param session: Specify a session. Otherwise, a session is created using the default profile.
        """
        self._session = session or Session()

    @property
    def session_id(self) -> str:
        """
        Retrieves the MCD API ID from the client's current session. For helping to identify requester client-side.
        """
        return self._session.id

    @property
    def session_name(self) -> str:
        """
        Retrieves the session name from the client's current session. For helping trace requests downstream.
        """
        return self._session.session_name

    @property
    def session_endpoint(self) -> str:
        """
        Retrieves the session MCD endpoint from the client's current session. By default, uses MCD_API_ENDPOINT.
        """
        return self._session.endpoint

    def _get_headers(self) -> Dict:
        """
        Gets headers from session for using the MCD API.

        Generates a trace ID to help trace (e.g. debug) specific requests downstream. Enable verbose logging to echo.
        """
        headers = {
            DEFAULT_MCD_API_ID_HEADER: self.session_id,
            DEFAULT_MCD_API_TOKEN_HEADER: self._session.token,
            DEFAULT_MCD_SESSION_ID: self.session_name,
            DEFAULT_MCD_TRACE_ID: str(uuid.uuid4()),
        }

        if self._session.user_id:
            headers[DEFAULT_MCD_USER_ID_HEADER] = self._session.user_id

        return headers

    def __call__(self,
                 query: Union[Query, Mutation, str],
                 variables: Optional[Dict] = None,
                 operation_name: Optional[str] = None,
                 retry_backoff: Optional[Backoff] = ExponentialBackoffJitter(DEFAULT_RETRY_INITIAL_WAIT_TIME,
                                                                             DEFAULT_RETRY_MAX_WAIT_TIME),
                 timeout_in_seconds: Optional[int] = 30) -> Union[Query, Mutation, Box, BoxList]:
        """
        Make a request to the MCD API.

        :param query: GraphQL query or mutation to execute. Can pass a string or Query/Mutation object.
        :param variables: Any variables to use with the query.
        :param operation_name: Name of the operation.
        :param retry_backoff: Set the retry backoff strategy. Defaults to an exponential backoff strategy with jitter.
        :param timeout_in_seconds: Set timeout of request. Requests cannot exceed 30 seconds.

        :return: Returns a Query or Mutation object with the response if the input query was a Query or Mutation object.
            If the input was a string a Box object containing the response is returned.
            Raises GqlError if any errors are found in the response.
            It will continually retry requests with errors using the provided `retry_backoff` parameter.

        Box is a transparent replacement for a dictionary - converting CamelCase to snake_case and allowing using dot
        notation in lookups. Can use .to_dict() to get a regular dictionary.
        """
        headers = self._get_headers()
        logger.info(f"Sending request to '{self.session_endpoint}' with trace ID '{headers[DEFAULT_MCD_TRACE_ID]}' "
                    f"in named session '{headers[DEFAULT_MCD_SESSION_ID]}'.")

        request = Endpoint(
            url=self.session_endpoint,
            base_headers=headers,
            timeout=timeout_in_seconds,
            retry_backoff=retry_backoff,
        )
        response = request(query, variables=variables, operation_name=operation_name)

        if not isinstance(query, str):
            return query + response
        return Box(response, camel_killer_box=True).data



import json
from typing import Optional, List, Tuple, Any, Union, Dict, Callable

import requests
from box import Box
from requests import HTTPError, ReadTimeout

from pycarlo.common import get_logger
from pycarlo.common.utils import chunks
from pycarlo.core import Client
from pycarlo.features.dbt.queries import IMPORT_DBT_MANIFEST, IMPORT_DBT_RUN_RESULTS

logger = get_logger(__name__)


class InvalidFileFormatException(Exception):
    pass


class DbtImporter:
    """
    Import dbt artifacts to Monte Carlo
    """

    def __init__(self,
                 mc_client: Optional[Client] = None,
                 print_func: Optional[Callable] = logger.info):
        self._mc_client = mc_client or Client()
        self._print_func = print_func

    def import_dbt_manifest(self,
                            dbt_manifest: Union[str, Dict],
                            project_name: Optional[str] = None,
                            batch_size: int = 10,
                            default_resource: Optional[str] = None) -> List[str]:
        """
        Import a dbt manifest

        :param dbt_manifest: either str indicating filename, or dict conforming to dbt manifest schema
        :param project_name: project_name to associate with manifest
        :param batch_size: import in batches of `batch_size` manifest elements
        :param default_resource: if account has multiple warehouses, define default_resource to choose
                                 the warehouse to associate with this dbt manifest. Can be either the
                                 warehouse's name or UUID

        :return: List of dbt node ID's that were successfully imported
        """
        if isinstance(dbt_manifest, str):
            with open(dbt_manifest, 'r') as f:
                dbt_manifest = Box(json.load(f))
        else:
            dbt_manifest = Box(dbt_manifest)

        try:
            dbt_schema_version = dbt_manifest.metadata.dbt_schema_version
            nodes = dbt_manifest.nodes
        except KeyError:
            raise InvalidFileFormatException("Unexpected format of input file. Ensure that input file is a valid DBT manifest.json file")

        node_items = list(nodes.items())
        self._print_func(f"\nImporting {len(node_items)} DBT objects into Monte Carlo catalog. please wait...")

        node_ids_imported = []
        all_bad_responses = []
        for nodes_items in chunks(node_items, batch_size):
            node_ids, bad_responses = self._do_make_request(dbt_schema_version, nodes_items, project_name,
                                                            default_resource=default_resource)
            if len(node_ids) > 0:
                self._print_func(f"Imported {len(node_ids)} DBT objects.")
            node_ids_imported.extend(node_ids)
            all_bad_responses.extend(bad_responses)

        if all_bad_responses:
            self._print_func("\nEncountered invalid responses.", all_bad_responses)

        self._print_func(f"\nImported a total of {len(node_ids_imported)} DBT objects into Monte Carlo catalog.\n")

        return node_ids_imported

    def _do_make_request(self, dbt_schema_version: str, nodes_items_list: List, project_name: Optional[str],
                         default_resource: Optional[str] = None) -> Tuple[List[str], List[Any]]:
        try:
            response = self._mc_client(
                query=IMPORT_DBT_MANIFEST,
                variables=dict(
                    dbtSchemaVersion=dbt_schema_version,
                    manifestNodesJson=json.dumps(dict(nodes_items_list)),
                    projectName=project_name,
                    defaultResource=default_resource
                )
            )

            try:
                return response.import_dbt_manifest.response.node_ids_imported, []
            except KeyError:
                return [], [response]

        except (HTTPError, ReadTimeout) as e:
            if isinstance(e, ReadTimeout) or \
                    (isinstance(e, HTTPError) and e.response.status_code == requests.codes.gateway_timeout):
                self._print_func(f"Import timed out with {e}, trying again with smaller batches.")

                if len(nodes_items_list) == 1:
                    raise RuntimeError("Could not split batch any further, exiting!")

                # Possible for the request to time out if there is too much data
                # Just send each one-by-one
                all_node_ids, all_bad_requests = [], []
                for single_nodes_items in chunks(nodes_items_list, 1):
                    node_ids, bad_requests = self._do_make_request(dbt_schema_version, single_nodes_items, project_name)
                    all_node_ids.extend(node_ids)
                    all_bad_requests.extend(all_bad_requests)

                return all_node_ids, all_bad_requests
            else:
                raise

    def import_run_results(self,
                           dbt_run_results: Union[str, Dict],
                           project_name: Optional[str] = None,
                           run_id: Optional[str] = None,
                           run_logs: Optional[str] = None) -> int:
        """
        Import dbt run results

        :param dbt_run_results: either str indicating filename, or dict conforming to dbt run results
        :param project_name: project_name to associate with run results (Optional)
        :param run_id: run_id to associate with run results (Optional)
        :param run_logs: dbt run log output to store with run (Optional)

        :return: number of run results imported
        """
        if isinstance(dbt_run_results, str):
            with open(dbt_run_results, 'r') as f:
                dbt_run_results = Box(json.load(f))
        else:
            dbt_run_results = Box(dbt_run_results)

        try:
            dbt_schema_version = dbt_run_results.metadata.dbt_schema_version
        except KeyError:
            raise InvalidFileFormatException(
                "Unexpected format of input file. Ensure that input file is a valid DBT run_results.json file")

        response = self._mc_client(
            query=IMPORT_DBT_RUN_RESULTS,
            variables=dict(
                dbtSchemaVersion=dbt_schema_version,
                runResultsJson=json.dumps(dbt_run_results),
                projectName=project_name,
                runId=run_id,
                runLogs=run_logs
            )
        )

        try:
            num_results_imported = response.import_dbt_run_results.response.num_results_imported
        except KeyError:
            num_results_imported = 0

        self._print_func(f"\nImported a total of {num_results_imported} DBT run results into Monte Carlo\n")

        return num_results_imported


import json
import os
import pathlib
from unittest import TestCase
from unittest.mock import Mock

import requests
from box import Box
from requests import HTTPError, Response

from pycarlo.features.dbt import DbtImporter
from pycarlo.features.dbt.queries import IMPORT_DBT_MANIFEST, IMPORT_DBT_RUN_RESULTS


class DbtImportServiceTest(TestCase):
    def test_import_dbt_manifest(self):
        self._client_mock = Mock(return_value=Box({
            'import_dbt_manifest': {
                'response': {
                    'node_ids_imported': [
                        "model.analytics.metric_types",
                        "model.analytics.recent_metrics",
                        "model.analytics.lineage_nodes"
                    ]
                }
            }
        }))

        service = DbtImporter(
            mc_client=self._client_mock
        )

        manifest_file = os.path.join(pathlib.Path(__file__).parent.resolve(), 'sample_manifest.json')
        node_ids_imported = service.import_dbt_manifest(manifest_file, default_resource='snowflake')

        with open(manifest_file, 'r') as f:
            dbt_manifest = Box(json.load(f))

        self._client_mock.assert_called_once_with(
            query=IMPORT_DBT_MANIFEST,
            variables=dict(
                dbtSchemaVersion='https://schemas.getdbt.com/dbt/manifest/v2.json',
                manifestNodesJson=json.dumps(dbt_manifest.nodes.to_dict()),
                projectName=None,
                defaultResource='snowflake'
            )
        )

        self.assertEqual(
            node_ids_imported,
            ['model.analytics.metric_types', 'model.analytics.recent_metrics', 'model.analytics.lineage_nodes']
        )

    def test_import_dbt_manifest_retry(self):
        def create_responses(*args, **kwargs):
            nodes = json.loads(kwargs['variables']['manifestNodesJson'])

            if len(list(nodes.items())) == 1:
                return Box({
                    'import_dbt_manifest': {
                        'response': {
                            'node_ids_imported': [list(nodes.keys())[0]]
                        }
                    }
                })

            response = Response()
            response.status_code = requests.codes.gateway_timeout

            raise HTTPError(response=response)

        self._client_mock = Mock(side_effect=create_responses)

        importer = DbtImporter(
            mc_client=self._client_mock
        )

        manifest_file = os.path.join(pathlib.Path(__file__).parent.resolve(), 'sample_manifest.json')
        node_ids_imported = importer.import_dbt_manifest(manifest_file)

        # client call will timeout, then will send each one-by-one
        # client will be called a total of 4 times
        self.assertEqual(4, self._client_mock.call_count)

        self.assertEqual(
            node_ids_imported,
            ['model.analytics.metric_types', 'model.analytics.recent_metrics', 'model.analytics.lineage_nodes']
        )

    def test_import_dbt_manifest_retry_bail_out(self):
        def create_responses(*args, **kwargs):
            response = Response()
            response.status_code = requests.codes.gateway_timeout

            raise HTTPError(response=response)

        self._client_mock = Mock(side_effect=create_responses)

        importer = DbtImporter(
            mc_client=self._client_mock
        )

        manifest_file = os.path.join(pathlib.Path(__file__).parent.resolve(), 'sample_manifest.json')

        with self.assertRaises(RuntimeError):
            node_ids_imported = importer.import_dbt_manifest(manifest_file)

        # make_request_v2() will always timeout
        # First request will time out
        # Then the next will time out, which has a batch size of 1, at which point it bails out
        self.assertEqual(2, self._client_mock.call_count)

    def test_import_dbt_run_results(self):
        self._client_mock = Mock(return_value=Box({
            'import_dbt_run_results': {
                'response': {
                    'num_results_imported': 4
                }
            }
        }))

        service = DbtImporter(
            mc_client=self._client_mock
        )

        run_results_file = os.path.join(pathlib.Path(__file__).parent.resolve(), 'sample_run_results.json')
        with open(run_results_file, 'r') as f:
            run_results = Box(json.load(f))

        node_ids_imported = service.import_run_results(run_results_file)

        self._client_mock.assert_called_once_with(
            query=IMPORT_DBT_RUN_RESULTS,
            variables=dict(
                dbtSchemaVersion='https://schemas.getdbt.com/dbt/run-results/v2.json',
                runResultsJson=json.dumps(run_results),
                projectName=None,
                runId=None,
                runLogs=None
            )
        )

        self.assertEqual(node_ids_imported, 4)


from unittest import TestCase
from unittest.mock import Mock

from pycarlo.features.dbt import DbtCloudClient
from pycarlo.features.dbt import DbtCloudImporter
from pycarlo.features.dbt import DbtImporter
from pycarlo.features.dbt.queries import CREATE_PROJECT


class DbtCloudImporterTest(TestCase):
    def test_import_dbt_cloud(self):
        dbt_cloud_client = Mock(autospec=DbtCloudClient)
        dbt_importer = Mock(autospec=DbtImporter)
        client_mock = Mock()

        dbt_cloud_importer = DbtCloudImporter(
            dbt_cloud_client=dbt_cloud_client,
            dbt_importer=dbt_importer,
            client=client_mock
        )

        dbt_cloud_client.get_projects.return_value = ({}, [
            {
                'id': 100,
                'name': 'my dbt project'
            }
        ])

        dbt_cloud_client.get_jobs.return_value = ({}, [
            {
                'id': 200,
                'name': 'my job'
            }
        ])

        dbt_cloud_client.get_runs.return_value = ({}, [
            {
                'id': 300,
                'is_complete': True,
                'project_id': 100,
                'run_steps': [
                    {
                        'logs': 'log 1'
                    },
                    {
                        'logs': 'log 2'
                    },
                ]
            }
        ])

        def mock_get_artifact(run_id: str, artifact_path: str):
            return { 'mock_artifact': artifact_path }
        dbt_cloud_client.get_run_artifact.side_effect = mock_get_artifact

        dbt_cloud_importer.import_dbt_cloud()

        # Should have created a project
        client_mock.assert_called_once_with(
            query=CREATE_PROJECT,
            variables=dict(
                projectName=100,
                source='DBT_CLOUD'
            )
        )

        dbt_importer.import_dbt_manifest.assert_called_once_with(
            dbt_manifest={
                'mock_artifact': 'manifest.json'
            },
            project_name=100,
            default_resource=None
        )

        dbt_importer.import_run_results.assert_called_once_with(
            dbt_run_results={
                'mock_artifact': 'run_results.json'
            },
            project_name=100,
            run_id=300,
            run_logs='log 1\nlog 2'
        )

        # Test with setting default_resource
        dbt_importer.reset_mock()
        dbt_cloud_importer.import_dbt_cloud(default_resource='snowflake')
        dbt_importer.import_dbt_manifest.assert_called_once_with(
            dbt_manifest={
                'mock_artifact': 'manifest.json'
            },
            project_name=100,
            default_resource='snowflake'
        )



import copy
import json
from unittest import TestCase
from unittest.mock import patch, Mock

from box import Box

from pycarlo.core import Client
from pycarlo.features.circuit_breakers import CircuitBreakerService
from pycarlo.features.circuit_breakers.exceptions import CircuitBreakerPollException
from pycarlo.features.exceptions import CircuitBreakerPipelineException

SAMPLE_BREACH_COUNT = 42
SAMPLE_RULE_ID = '1234'
SAMPLE_JOB_EXECUTION = '5678'
SAMPLE_TRIGGER = Box({'trigger_circuit_breaker_rule': {'job_execution_uuid': SAMPLE_JOB_EXECUTION}})
SAMPLE_BREACH = Box({'payload': {'breach_count': SAMPLE_BREACH_COUNT}}, default_box_attr=None, default_box=True)
SAMPLE_ERROR = Box(
    {'payload': {'error': "Exception('this is a test')", 'traceback': ['Line 10']}},
    default_box_attr=None,
    default_box=True
)


class MockSate:
    get_circuit_breaker_rule_state = Box(
        {'status': 'PROCESSING_COMPLETE', 'log': json.dumps([SAMPLE_BREACH])},
        default_box_attr=None,
        default_box=True
    )


class CircuitBreakerServiceTest(TestCase):
    def setUp(self) -> None:
        self._client_mock = Mock(autospec=Client)

        self._service = CircuitBreakerService(mc_client=self._client_mock)

    @patch.object(CircuitBreakerService, 'poll')
    def test_trigger_and_poll_in_breach(self, poll_mock):
        poll_mock.return_value = SAMPLE_BREACH_COUNT
        self.assertTrue(self._test_trigger_and_poll(poll_mock=poll_mock))

    @patch.object(CircuitBreakerService, 'poll')
    def test_trigger_and_poll_no_breach(self, poll_mock):
        poll_mock.return_value = 0
        self.assertFalse(self._test_trigger_and_poll(poll_mock=poll_mock))

    @patch.object(CircuitBreakerService, 'trigger')
    def _test_trigger_and_poll(self, trigger_mock, poll_mock) -> bool:
        trigger_mock.return_value = SAMPLE_JOB_EXECUTION

        in_breach = self._service.trigger_and_poll(rule_uuid=SAMPLE_RULE_ID)
        trigger_mock.assert_called_once_with(rule_uuid=SAMPLE_RULE_ID)
        poll_mock.assert_called_once_with(job_execution_uuid=SAMPLE_JOB_EXECUTION, timeout_in_minutes=5)

        return in_breach

    def test_trigger(self):
        self._client_mock.return_value = SAMPLE_TRIGGER

        self.assertEqual(self._service.trigger(rule_uuid=SAMPLE_RULE_ID), SAMPLE_JOB_EXECUTION)
        self._client_mock.assert_called_once()
        self.assertEqual(
            self._trim_whitespace(self._client_mock.call_args[0][0]),
            self._trim_whitespace(
                f'mutation {{triggerCircuitBreakerRule(ruleUuid: "{SAMPLE_RULE_ID}") {{jobExecutionUuid}}}}')
        )

    @patch.object(CircuitBreakerService, '_poll')
    def test_poll_in_breach(self, internal_poll_mock):
        internal_poll_mock.return_value = SAMPLE_BREACH
        self.assertEqual(self._service.poll(job_execution_uuid=SAMPLE_JOB_EXECUTION), SAMPLE_BREACH_COUNT)
        internal_poll_mock.assert_called_once_with(job_execution_uuid=SAMPLE_JOB_EXECUTION, timeout_in_minutes=5)

    @patch.object(CircuitBreakerService, '_poll')
    def test_poll_not_in_breach(self, internal_poll_mock):
        log = copy.deepcopy(SAMPLE_BREACH)
        log.payload.breach_count = 0

        internal_poll_mock.return_value = log
        self.assertEqual(self._service.poll(job_execution_uuid=SAMPLE_JOB_EXECUTION), 0)
        internal_poll_mock.assert_called_once_with(job_execution_uuid=SAMPLE_JOB_EXECUTION, timeout_in_minutes=5)

    @patch.object(CircuitBreakerService, '_poll')
    def test_poll_with_pipeline_error(self, internal_poll_mock):
        internal_poll_mock.return_value = SAMPLE_ERROR
        with self.assertRaises(CircuitBreakerPipelineException) as context:
            self._service.poll(job_execution_uuid=SAMPLE_JOB_EXECUTION)
        self.assertEqual(str(context.exception), f'Execution pipeline errored out. Details:\n{SAMPLE_ERROR}')

    @patch.object(CircuitBreakerService, '_poll')
    def test_poll_with_timeout(self, internal_poll_mock):
        internal_poll_mock.return_value = None
        with self.assertRaises(CircuitBreakerPollException) as context:
            self._service.poll(job_execution_uuid=SAMPLE_JOB_EXECUTION)
        self.assertEqual(str(context.exception), 'Polling timed out or contains a malformed log.')

    def test_internal_poll(self):
        self._client_mock.return_value = MockSate
        self.assertEqual(
            self._service._poll(job_execution_uuid=SAMPLE_JOB_EXECUTION, timeout_in_minutes=10),
            SAMPLE_BREACH
        )

    @patch('pycarlo.features.circuit_breakers.service.time')
    def test_internal_poll_with_timeout(self, mock_time):
        mock_time.time.side_effect = [1643751961, 1643752021, 1644356761]
        self.assertIsNone(self._service._poll(job_execution_uuid=SAMPLE_JOB_EXECUTION, timeout_in_minutes=10))
        mock_time.sleep.assert_called_once_with(15)

    @staticmethod
    def _trim_whitespace(_str: str) -> str:
        return ''.join(str(_str).split())



from random import randrange


def downtime() -> None:
    """
    A bit of fun. Demos SDK feature extensions.
    """
    if randrange(10) >= 5:
        raise SystemExit('Bad luck.')
    print('No data downtime found.')



from typing import Optional, Callable

from pycarlo.common import settings, get_logger
from pycarlo.core import Client
from pycarlo.features.dbt.dbt_importer import DbtImporter
from pycarlo.features.dbt.dbt_cloud_client import DbtCloudClient
from pycarlo.features.dbt.queries import CREATE_PROJECT

_ARTIFACT_MANIFEST = 'manifest.json'
_ARTIFACT_RUN_RESULTS = 'run_results.json'

logger = get_logger(__name__)


class DbtCloudImporter:
    """
    Import manifest and run results from dbt cloud to Monte Carlo
    """
    def __init__(self,
                 dbt_cloud_client: Optional[DbtCloudClient] = None,
                 client: Optional[Client] = None,
                 dbt_importer: Optional[DbtImporter] = None,
                 print_func: Optional[Callable] = logger.info):
        self._client = client or Client()
        self._print_func = print_func
        self._dbt_importer = dbt_importer or DbtImporter(
            mc_client=self._client,
            print_func=print_func
        )
        self._dbt_client = dbt_cloud_client or DbtCloudClient()

    def import_dbt_cloud(self,
                         project_id: Optional[str] = None,
                         job_id: Optional[str] = None,
                         manifest_only: Optional[bool] = False,
                         default_resource: Optional[str] = None):
        """
        Use dbt API to gather all projects.
        For each project, gather all jobs.
        For each job, retrieve the latest run.
        For each latest run, retrieve manifest.json and run_results.json artifacts, and import to MC
        """
        if project_id:
            status, project = self._dbt_client.get_project(project_id)
            projects = [project]
        else:
            status, projects = self._dbt_client.get_projects()

        all_jobs = []
        if job_id:
            status, job = self._dbt_client.get_job(job_id)
            all_jobs.append((job['id'], job['name']))
        else:
            for project in projects:
                name, project_id = project['name'], project['id']
                self._print_func(f'Project: {name} ({project_id})')

                self._create_dbt_project(project_id)

                # Get all jobs in project
                status, jobs = self._dbt_client.get_jobs(project_id)

                for job in jobs:
                    job_name, job_id = job['name'], job['id']
                    self._print_func(f'* Found Job: {job_name} ({job_id})')
                    all_jobs.append((job_id, job_name))

        self._print_func('')

        for job_id, job_name in all_jobs:
            self._print_func(f'Processing job: {job_name} ({job_id})')
            status, runs = self._dbt_client.get_runs(job_definition_id=job_id)
            if not runs:
                self._print_func(f'No runs found for job')
                continue

            # Get first run that is complete
            for run in runs:
                if run['is_complete']:
                    break
            run_id = run['id']

            try:
                run_steps = run.get('run_steps', [])
                run_logs = '\n'.join([step['logs'] for step in run_steps])

                run_results = self._dbt_client.get_run_artifact(run_id=run_id, artifact_path=_ARTIFACT_RUN_RESULTS)
                manifest = self._dbt_client.get_run_artifact(run_id=run_id, artifact_path=_ARTIFACT_MANIFEST)

                self._print_func("=================================================")
                self._print_func(f"Importing manifest for run id={run_id}")
                self._print_func("=================================================")
                self._dbt_importer.import_dbt_manifest(
                    dbt_manifest=manifest,
                    project_name=run['project_id'],
                    default_resource=default_resource
                )

                if not manifest_only:
                    self._print_func("=================================================")
                    self._print_func(f"Importing run results for run id={run_id}")
                    self._print_func("=================================================")
                    self._dbt_importer.import_run_results(
                        dbt_run_results=run_results,
                        project_name=run['project_id'],
                        run_id=run['id'],
                        run_logs=run_logs
                    )
            except Exception as e:
                self._print_func(f'Could not import data from run, id={run_id}, reason={e}')

    def _create_dbt_project(self, project_id: str):
        self._client(
            query=CREATE_PROJECT,
            variables=dict(
                projectName=project_id,
                source='DBT_CLOUD'
            )
        )


from unittest import TestCase
from unittest.mock import patch, Mock

from box import Box
from requests.exceptions import Timeout
from responses import activate, mock, post, registries

from pycarlo.common.errors import GqlError
from pycarlo.common.retries import Backoff
from pycarlo.common.settings import DEFAULT_MCD_API_ENDPOINT
from pycarlo.core import Session, Client, Query

MCD_API_ENDPOINT = DEFAULT_MCD_API_ENDPOINT

MOCK_MCD_ID = 'foo'
MOCK_MCD_TOKEN = 'bar'
MOCK_SESSION_ID = 'qux'
MOCK_TRACE_ID = 'baz'
MOCK_USER_ID = 'uid'
MOCK_HEADERS = {
    'x-mcd-id': MOCK_MCD_ID,
    'x-mcd-session-id': MOCK_SESSION_ID,
    'x-mcd-token': MOCK_MCD_TOKEN,
    'x-mcd-trace-id': MOCK_TRACE_ID,
    'user-id': MOCK_USER_ID
}
MOCK_TIMEOUT = 16
MOCK_EMAIL = 'dresden@montecarlodata.com'
MOCK_GET_USER_QUERY = """
query {
  getUser {
    email
  }
}
"""
MOCK_GOOD_RESPONSE = {'data': {'getUser': {'email': MOCK_EMAIL}}, 'errors': []}




class MockBackoff(Backoff):
    def __init__(self, count: int):
        super(MockBackoff, self).__init__(start=0, maximum=0)
        self.count = count

    def backoff(self, attempt: int) -> float:
        return 0

    def delays(self):
        for i in range(self.count):
            yield 0


class ClientTest(TestCase):
    def setUp(self) -> None:
        self._session = Mock(autospec=Session)
        self._session.endpoint = MCD_API_ENDPOINT
        self._backoff = MockBackoff(count=0)
        self._client = Client(session=self._session)

    def test_get_session_id(self):
        self._session.id = MOCK_MCD_ID
        self.assertEqual(self._client.session_id, MOCK_MCD_ID)

    def test_get_session_name(self):
        self._session.session_name = MOCK_SESSION_ID
        self.assertEqual(self._client.session_name, MOCK_SESSION_ID)

    def test_get_session_endpoint(self):
        endpoint = 'test.com'
        self._session.endpoint = endpoint
        self.assertEqual(self._client.session_endpoint, endpoint)

    @patch('pycarlo.core.client.uuid')
    def test_get_headers(self, mock_uuid):
        self._session.id = MOCK_MCD_ID
        self._session.token = MOCK_MCD_TOKEN
        self._session.session_name = MOCK_SESSION_ID
        self._session.user_id = MOCK_USER_ID

        mock_uuid.uuid4.return_value = MOCK_TRACE_ID

        self.assertEqual(self._client._get_headers(), MOCK_HEADERS)

    @patch.object(Client, '_get_headers', lambda *args: MOCK_HEADERS)
    @activate(registry=registries.OrderedRegistry)
    def test_call(self):
        post(MCD_API_ENDPOINT, json=MOCK_GOOD_RESPONSE, status=200)
        self.assertEqual(
            self._client(query=MOCK_GET_USER_QUERY, retry_backoff=self._backoff, timeout_in_seconds=MOCK_TIMEOUT),
            Box(MOCK_GOOD_RESPONSE, camel_killer_box=True).data
        )
        mock.assert_call_count(MCD_API_ENDPOINT, 1)

    @patch.object(Client, '_get_headers', lambda *args: MOCK_HEADERS)
    @activate(registry=registries.OrderedRegistry)
    def test_call_with_object(self):
        post(MCD_API_ENDPOINT, json=MOCK_GOOD_RESPONSE, status=200)

        query = Query()
        query.get_user.__fields__('email')

        response = self._client(query=query, retry_backoff=self._backoff, timeout_in_seconds=MOCK_TIMEOUT)
        self.assertEqual(response.get_user.email, MOCK_EMAIL)
        mock.assert_call_count(MCD_API_ENDPOINT, 1)

    @patch.object(Client, '_get_headers', lambda *args: MOCK_HEADERS)
    @activate(registry=registries.OrderedRegistry)
    def test_call_with_single_gql_error(self):
        post(
            MCD_API_ENDPOINT,
            json={'data': None, 'errors': [{'message': 'Life is a journey. Time is a river. The door is ajar.'}]},
            status=200,
        )

        with self.assertRaises(GqlError) as err:
            self._client(query=MOCK_GET_USER_QUERY, retry_backoff=self._backoff, timeout_in_seconds=MOCK_TIMEOUT)
        self.assertEqual('Life is a journey. Time is a river. The door is ajar.', err.exception.message)
        mock.assert_call_count(MCD_API_ENDPOINT, 1)

    @patch.object(Client, '_get_headers', lambda *args: MOCK_HEADERS)
    @activate(registry=registries.OrderedRegistry)
    def test_call_with_multiple_gql_errors(self):
        post(
            MCD_API_ENDPOINT,
            json={'data': None, 'errors': [{'message': 'Life is a journey.'}, {'message': 'Time is a river.'}]},
            status=200,
        )

        with self.assertRaises(GqlError) as err:
            self._client(query=MOCK_GET_USER_QUERY, retry_backoff=self._backoff, timeout_in_seconds=MOCK_TIMEOUT)
        self.assertEqual('Life is a journey.\nTime is a river.', err.exception.message)
        mock.assert_call_count(MCD_API_ENDPOINT, 1)

    @patch.object(Client, '_get_headers', lambda *args: MOCK_HEADERS)
    @activate(registry=registries.OrderedRegistry)
    def test_call_with_user_errors(self):
        post(
            MCD_API_ENDPOINT,
            json={'message': 'Invalid field value', 'field': 'name'},
            status=400,
            headers={'X-Custom': 'blah blah'},
        )

        backoff = MockBackoff(2)
        with self.assertRaises(GqlError) as err:
            self._client(query=MOCK_GET_USER_QUERY, retry_backoff=backoff, timeout_in_seconds=MOCK_TIMEOUT)
        self.assertEqual(
            {'Content-Type': 'application/json', 'X-Custom': 'blah blah'},
            err.exception.headers,
        )
        self.assertEqual(
            400,
            err.exception.status_code,
        )
        self.assertEqual(
            'Invalid field value',
            err.exception.message,
        )
        self.assertEqual(
            '{"message": "Invalid field value", "field": "name"}',
            err.exception.body,
        )
        self.assertEqual(
            f'400 Client Error: Bad Request for url: {MCD_API_ENDPOINT}',
            str(err.exception),
        )
        mock.assert_call_count(MCD_API_ENDPOINT, 1)

    @patch.object(Client, '_get_headers', lambda *args: MOCK_HEADERS)
    @activate(registry=registries.OrderedRegistry)
    def test_call_with_network_errors_no_retries(self):
        post(MCD_API_ENDPOINT, body='Oops', status=503)

        with self.assertRaises(GqlError) as err:
            self._client(query=MOCK_GET_USER_QUERY, retry_backoff=self._backoff, timeout_in_seconds=MOCK_TIMEOUT)
        self.assertEqual(
            {'Content-Type': 'text/plain'},
            err.exception.headers,
        )
        self.assertEqual(
            503,
            err.exception.status_code,
        )
        self.assertEqual(
            'Oops',
            err.exception.message,
        )
        self.assertEqual(
            'Oops',
            err.exception.body,
        )
        self.assertEqual(
            f'503 Server Error: Service Unavailable for url: {MCD_API_ENDPOINT}',
            str(err.exception),
        )
        mock.assert_call_count(MCD_API_ENDPOINT, 1)

    @patch.object(Client, '_get_headers', lambda *args: MOCK_HEADERS)
    @activate(registry=registries.OrderedRegistry)
    def test_call_with_network_errors_retried_with_success(self):
        post(MCD_API_ENDPOINT, body='Oops', status=503)
        post(MCD_API_ENDPOINT, body='Oops', status=503)
        post(MCD_API_ENDPOINT, json=MOCK_GOOD_RESPONSE, status=200)

        backoff = MockBackoff(2)
        response = self._client(query=MOCK_GET_USER_QUERY, retry_backoff=backoff, timeout_in_seconds=MOCK_TIMEOUT)
        self.assertEqual(
            MOCK_EMAIL,
            response.get_user.email,
        )
        mock.assert_call_count(MCD_API_ENDPOINT, 3)

    @patch.object(Client, '_get_headers', lambda *args: MOCK_HEADERS)
    @activate(registry=registries.OrderedRegistry)
    def test_call_with_network_errors_retried_to_exhaustion(self):
        post(MCD_API_ENDPOINT, body='Oops', status=503)
        post(MCD_API_ENDPOINT, body='Oops', status=503)
        post(MCD_API_ENDPOINT, body='Oops', status=503)
        post(MCD_API_ENDPOINT, body='Oops', status=503)
        post(MCD_API_ENDPOINT, body='Oops', status=503)
        post(MCD_API_ENDPOINT, body='Oops', status=503)

        backoff = MockBackoff(4)
        with self.assertRaises(GqlError) as err:
            self._client(query=MOCK_GET_USER_QUERY, retry_backoff=backoff, timeout_in_seconds=MOCK_TIMEOUT)
        self.assertEqual(
            {'Content-Type': 'text/plain'},
            err.exception.headers,
        )
        self.assertEqual(
            503,
            err.exception.status_code,
        )
        self.assertEqual(
            'Oops',
            err.exception.message,
        )
        self.assertEqual(
            'Oops',
            err.exception.body,
        )
        self.assertEqual(
            f'503 Server Error: Service Unavailable for url: {MCD_API_ENDPOINT}',
            str(err.exception),
        )
        mock.assert_call_count(MCD_API_ENDPOINT, 5)

    @patch.object(Client, '_get_headers', lambda *args: MOCK_HEADERS)
    @activate(registry=registries.OrderedRegistry)
    def test_call_with_timeout_retried_with_success(self):
        post(MCD_API_ENDPOINT, body=Timeout())
        post(MCD_API_ENDPOINT, body=Timeout())
        post(MCD_API_ENDPOINT, json=MOCK_GOOD_RESPONSE, status=200)

        backoff = MockBackoff(2)
        response = self._client(query=MOCK_GET_USER_QUERY, retry_backoff=backoff, timeout_in_seconds=MOCK_TIMEOUT)
        self.assertEqual(
            MOCK_EMAIL,
            response.get_user.email,
        )
        mock.assert_call_count(MCD_API_ENDPOINT, 3)

    @patch.object(Client, '_get_headers', lambda *args: MOCK_HEADERS)
    @activate(registry=registries.OrderedRegistry)
    def test_call_with_timeout_retried_to_exhaustion(self):
        post(MCD_API_ENDPOINT, body=Timeout())
        post(MCD_API_ENDPOINT, body=Timeout())
        post(MCD_API_ENDPOINT, body=Timeout())
        post(MCD_API_ENDPOINT, body=Timeout())
        post(MCD_API_ENDPOINT, body=Timeout())
        post(MCD_API_ENDPOINT, body=Timeout())

        backoff = MockBackoff(4)
        with self.assertRaises(expected_exception=(GqlError, Timeout)) as err:
            self._client(query=MOCK_GET_USER_QUERY, retry_backoff=backoff, timeout_in_seconds=MOCK_TIMEOUT)
        self.assertIsInstance(err.exception, Timeout)
        mock.assert_call_count(MCD_API_ENDPOINT, 5)
